#perl

#
# Title: Texel Density Kit
# Version: 1.0
# Author: James O'Hare
# http://www.farfarer.com
# jamesohare@gmail.com
#

#
# Description:
# Tool to get/set texel density of selected polygons or UV islands.
#

#
# Note:
# A big thanks to Muhamed Toromanovic for his excellent kit tutorial, it's been a big help in putting this into kit form.
# http://www.toroman.de/modo/tutorials/creating-a-kit-precision-scale/
#

#
# Installation:
# Put the folder containing this file and the config file into your Scripts directory.
# Restart modo.
# You should now see it appended to the UV menu inside modo.
#

#
# Usage:
# @texden arguments
# 
# Arguments			- Function
#
# get uv			- Get the average texel density of the UV islands of the selected polygons.
# get poly			- Get the average texel density of the selected polygons.
#
# set uv each		- Scale the UV island of each selected polygon so each one's average matches the current texel density.
# set uv all		- Scale all UV islands of each selected polygon so their average matches the current texel density.
#
# set poly each		- Scale each selected polygon so each one's average matchs the current texel density.
# set poly all		- Scale all selected polygon so their average matches the current texel density.
#
# gettexturesize	- Get the largest axis of the currently selected texture.
# getrendersize		- Get the largest axis of the render output.
#

#______________________________________________________________________________________________________ Helper Functions

# Get vector length.
sub vector_length {
	my @vector = @_;
	my $length = sqrt(($vector[0] * $vector[0]) + ($vector[1] * $vector[1]) + ($vector[2] * $vector[2]));
	return $length;
}

# Subtract vector B from vector A.
sub vector_subtract {
	my @vectorA = @{ $_[0] };
	my @vectorB = @{ $_[1] };

	my @vectorC;
	$vectorC[0] = $vectorA[0] - $vectorB[0];
	$vectorC[1] = $vectorA[1] - $vectorB[1];
	$vectorC[2] = $vectorA[2] - $vectorB[2];

	return @vectorC;
 }

# Get the area of the triangle with points point A, B and C.
sub triangle_area {
	my @pointA = @{ $_[0] };
	my @pointB = @{ $_[1] };
	my @pointC = @{ $_[2] };

	my @edgeA = &vector_subtract(\@pointB, \@pointA);
	my @edgeB = &vector_subtract(\@pointC, \@pointB);
	my @edgeC = &vector_subtract(\@pointA, \@pointC);

	my $sideA = &vector_length(@edgeA);
	my $sideB = &vector_length(@edgeB);
	my $sideC = &vector_length(@edgeC);

	my $perim = ($sideA + $sideB + $sideC) / 2;

	my $triangleArea = sqrt($perim * ($perim - $sideA) * ($perim - $sideB) * ($perim - $sideC) );

 	return $triangleArea;
}

my $tempSetPolysChecked = "polysToBeProcessed";
my $tempSetPoly = "polyPrevSelection";
my $restorePolys = 0;

# Create temporary selection sets for selected elements.
sub saveSelection {
	# Polys
	#lxout("Storing poly selection.");
	lx("select.type polygon");
	if (lxq("query layerservice poly.N ? selected")) {
		lx("select.editSet $tempSetPoly add");
		$restorePolys = 1;
	}
}

# Select then remove temporary selection set previously selected elements.
sub restoreSelection {
	# Polys
	#lxout("Restoring poly selection.");
	lx("select.drop polygon");
	lx("select.type polygon");
	if ($restorePolys) {
		lx("!!select.useSet $tempSetPoly select");
		lx("!!select.deleteSet $tempSetPoly");
	}
}

# TEXTURE SIZE

# Find the currently selected clip in the shader tree, and then pull it's width and height.
sub getTextureSize() {
	my $textureSize = lxq("user.value texeldensity.imagesize ?");

	my $selectedTexture = lxq("texture.setIMap name:?");

	my @clips = lxq("query layerservice clips ? all");
	foreach $clip (@clips) {
		my $clipName = lxq("query layerservice clip.name ? $clip");

		if ( $clipName eq $selectedTexture ) {
			my $clipInfo = lxq("query layerservice clip.info ? $texClip");

			my @clipInfoArray = split(/ /, $clipInfo);
			my @clipWidthArray = split(/:/, $clipInfoArray[1]);
			my @clipHeightArray = split(/:/, $clipInfoArray[2]);
			my $clipHeight = $clipHeightArray[1];
			my $clipWidth = $clipWidthArray[1];

			#lxout("Clip: $clipName has width $clipWidth and height $clipHeight");

			if ($clipWidth > $clipHeight ) {
				$textureSize = $clipWidth;
			}
			else {
				$textureSize = $clipHeight;
			}
			last;
		}
	}

	lx("user.value texeldensity.imagesize $textureSize");
}

# Pull render output's width and height.
sub getRenderSize() {
	my $textureSize = lxq("user.value texeldensity.imagesize ?");

	my $renderWidth = lxq("render.res 0 ?");
	my $renderHeight = lxq("render.res 1 ?");

	if ($renderWidth > $renderHeight ) {
		$textureSize = $renderWidth;
	}
	else {
		$textureSize = $renderHeight;
	}

	lx("user.value texeldensity.imagesize $textureSize");
	lx("user.value texeldensity.imagesizeupdate false");
}

# IN:
# bbox : array			Bounds; min U, max U, min V, max V.
# uvPos : array			UV positions as returned straight from modo.
#
# OUT:
# bbox : array			Bounds checked against UV positions and updated if changed.
sub expandBBox() {

	my @bbox = @{ $_[0] };
	my @uvPos = @{ $_[1] };

	my $numVerts = scalar(@uvPos);
	my $i = 0;
	my $j = 1;
	while ($i < $numVerts) {
		$bbox[0] = $uvPos[$i] if $uvPos[$i] < $bbox[0];	# Min U
		$bbox[1] = $uvPos[$i] if $uvPos[$i] > $bbox[1]; # Max U
		$bbox[2] = $uvPos[$j] if $uvPos[$j] < $bbox[2]; # Min V
		$bbox[3] = $uvPos[$j] if $uvPos[$j] > $bbox[3]; # Min V
		$i += 2;
		$j += 2;
	}

	return @bbox;
}

# IN:
# uv : boolean				Whether to select the UV islands of selected polygon(s).
# bbox : boolean			Whether to get a bounding box for the UVs.
#
# OUT:
# texelDensity : float		Average texel density of selection.
sub getTexelDensity() {

	&quickChecks();

	# Bounding box check for action centre. Is there a better way to do this?
	my @bbox;
	if ($_[1]) {
		@bbox = ( 99999.0, -99999.0, 99999.0, -99999.0 );
	}

	my $UV3DRatio = 0.0;

	if ($_[0]) {
		# Get UV island of selected polygons.
		lx("select.polygonConnect uv");
	}

	my @polygons = lxq("query layerservice polys ? selected");
	my $numPolys = scalar(@polygons);

	foreach my $poly (@polygons) {

		# Get vertices of polygon, count them and find out triangle count.

		my @polyVerts = lxq("query layerservice poly.vertList ? $poly");
		my $vertCount = scalar(@polyVerts);
		my $triCount = $vertCount - 2;



		# 3D AREA

		my $poly3DArea = 0.0;
		my $i = 0;
		my @vert1Pos = lxq("query layerservice vert.pos ? $polyVerts[0]");
		while ($i < $triCount) {
			my @vert2Pos = lxq("query layerservice vert.pos ? $polyVerts[$i+1]");
			my @vert3Pos = lxq("query layerservice vert.pos ? $polyVerts[$i+2]");
			my $tri3DArea = &triangle_area(\@vert1Pos, \@vert2Pos, \@vert3Pos);
			$poly3DArea += $tri3DArea;
			$i++;
		}
		##lxout("3D area of polygon $poly is $poly3DArea");

		# UV AREA

		my @polyUVPos = lxq("query layerservice poly.vmapValue ? $poly");
		my $polyUVArea = 0.0;
		my $j = 0;
		my @uv1Pos = ("$polyUVPos[0]", "$polyUVPos[1]", "0.0");

		# Bounding box check.
		if ($_[1]) {
			@bbox = &expandBBox(\@bbox, \@polyUVPos);
		}

		while ($j < $triCount) {
			my @uv2Pos = ("$polyUVPos[$j*2+2]", "$polyUVPos[$j*2+3]", "0.0");
			my @uv3Pos = ("$polyUVPos[$j*2+4]", "$polyUVPos[$j*2+5]", "0.0");
			my $triUVArea = &triangle_area(\@uv1Pos, \@uv2Pos, \@uv3Pos);
			$polyUVArea += $triUVArea;
			$j++;
		}
		##lxout("UV area of polygon $poly is $polyUVArea");

		my $polyUV3DRatio = $polyUVArea / $poly3DArea;
		$UV3DRatio += $polyUV3DRatio;
	}



	# TEXEL DENSITY

	$UV3DRatio = $UV3DRatio / $numPolys;
	#lxout("3D:UV Ratio is $UV3DRatio");

	my $texelDensity = lxq("user.value texeldensity.imagesize ?") * $UV3DRatio;
	#lxout("Texel density is $texelDensity");

	if ($_[1]) {
		my $bbox_U = ($bbox[0] + $bbox[1]) / 2.0;
		my $bbox_V = ($bbox[2] + $bbox[3]) / 2.0;
		my @result = ( $texelDensity, $bbox_U, $bbox_V );
		return @result;
	}
	else {
		return $texelDensity;
	}
}

# IN:
# uv : boolean				Whether to select the UV islands of selected polygon(s).
sub setTexelDensity() {

	my @result = &getTexelDensity($_[0], 1);

	my $scaleValue = 1.0; # Default here, to be safe.
	if ($result[0] != 0.0) {
		$scaleValue = sqrt(lxq("user.value texeldensity.current ?") / $result[0]);
	}
	#lxout("Scale factor is $scaleValue");
	lx("tool.set TransformScale on");
	lx("tool.xfrmDisco false");
	lx("tool.viewType uv");
	lx("tool.set axis.origin on");
	lx("tool.set center.auto on");
	lx("tool.setAttr center.auto cenU $result[1]");
	lx("tool.setAttr center.auto cenV $result[2]");
	lx("tool.setAttr xfrm.transform SX $scaleValue");
	lx("tool.setAttr xfrm.transform SY $scaleValue");
	lx("tool.doApply");
	lx("tool.set TransformScale flush 0");
}

# IN:
# uv : boolean				Whether to select the UV islands of selected polygon(s).
sub setTexelDensityEach() {
	# Create a selection set containting the polygons we want to resize.
	lx("select.editSet $tempSetPolysChecked add");

	# Grab the main layer ID. We'll select this first later on to ensure that it remains the main layer.
	my $mainLayer = lxq("query layerservice layers ? main");
	my $mainLayerID = lxq("query layerservice layer.id ? $mainLayer");

	my @layers = lxq("query layerservice layers ? fg");
	foreach $layer (@layers) {
		my $layerID = lxq("query layerservice layer.id ? $layer");
		my $layerName = lxq("query layerservice layer.name ? $layer");
		lx("select.subItem $layerID set mesh");

		# Select the "to check" selection set.
		lx("select.drop polygon");
		lx("!!select.useSet $tempSetPolysChecked select");

		my $polysLeftToCheck = lxq("query layerservice poly.N ? selected");
		#lxout("Polys to check on layer $layer: $polysLeftToCheck");
		while ($polysLeftToCheck) {
			# Get selection. Pick first poly from selection.
			my @selectedPolygons = lxq("query layerservice polys ? selected");
			#lxout("Selected polys: @selectedPolygons");

			# Grab the first one from the selection.
			lx("select.element $layer polygon set $selectedPolygons[0] 0 0");
			&setTexelDensity($_[0], 1);

			# Remove current selection from selection set.
			lx("select.editSet $tempSetPolysChecked remove {}");
			# Select it again.
			lx("select.drop polygon");
			lx("!!select.useSet $tempSetPolysChecked select");

			# If there are no polys selected, we're done.
			$polysLeftToCheck = lxq("query layerservice poly.N ? selected");
			#lxout("Polys left to check on layer $layer: $polysLeftToCheck");
		}
	}

	# Remove the selection set.
	lx("!!select.deleteSet $tempSetPolysChecked");

	# Restore original layer selection. Main layer first.
	lx("select.subItem $mainLayerID set mesh");
	foreach my $layer (@layers) {
		my $layerID = lxq("query layerservice layer.id ? $layer");
		lx("select.subItem $layerID add mesh");
	}
}


#__________________________________________________________________________________________________________ Early Checks

sub quickChecks() {
	# Are any polys selected?
	if (!lxq("query layerservice poly.N ? selected")) {
		#lxout("No polygons selected.");
		return;
	}

	# Is UV map selected?
	my $uvMap = -1;
	my @vertMaps = lxq("query layerservice vmaps ? selected");
	foreach my $vertMap (@vertMaps) {
		my $vertMapType = lxq("query layerservice vmap.type ? $vertMap");
		if ($vertMapType eq "texture") {
			$uvMap = $vertMap;
			last;
		}
	}
	if ($uvMap == -1) {
		#lxout("No UV map selected.");
		return;
	}
}



#_________________________________________________________________________________________________________ Script Itself

if ((lxq("user.value texeldensity.imagesizeupdate ?")) or ($ARGV[0] eq "gettexturesize")) {
	&getTextureSize();
}

if ($ARGV[0] eq "get") {
	&saveSelection();

	my $texelDensity = lxq("user.value texeldensity.current ?");
	if ($ARGV[1] eq "uv") {
		$texelDensity = &getTexelDensity(1, 0);
	}
	elsif ($ARGV[1] eq "poly") {
		$texelDensity = &getTexelDensity(0, 0);
	}
	lx("user.value texeldensity.current $texelDensity");

	&restoreSelection();
}
elsif ($ARGV[0] eq "set") {
	&saveSelection();

	if ($ARGV[1] eq "uv") {
		if ($ARGV[2] eq "all") {
			&setTexelDensity(1);
		}
		elsif ($ARGV[2] eq "each") {
			&setTexelDensityEach(1);
		}
	}
	elsif ($ARGV[1] eq "poly") {
		if ($ARGV[2] eq "all") {
			&setTexelDensity(0);
		}
		elsif ($ARGV[2] eq "each") {
			&setTexelDensityEach(0);
		}
	}

	&restoreSelection();
}
elsif ($ARGV[0] eq "getrendersize") {
	&getRenderSize();
}